import fs from "fs";
import npath from "path";
import { fileURLToPath } from "url";
import { debug, error, info } from "../../core/logger.js";
import { prependForwardSlash } from "../../core/path.js";
import { BEFORE_HYDRATION_SCRIPT_ID } from "../../vite-plugin-scripts/index.js";
import { call as callEndpoint } from "../endpoint/index.js";
import { render } from "../render/core.js";
import { createLinkStylesheetElementSet, createModuleScriptElementWithSrcSet } from "../render/ssr-element.js";
import { getOutFile, getOutRoot, getOutFolder, getServerRoot } from "./common.js";
import { getPageDataByComponent, eachPageData } from "./internal.js";
import { bgMagenta, black, cyan, dim, magenta } from "kleur/colors";
import { getTimeStat } from "./util.js";
const MAX_CONCURRENT_RENDERS = 1;
function* throttle(max, inPaths) {
  let tmp = [];
  let i = 0;
  for (let path of inPaths) {
    tmp.push(path);
    if (i === max) {
      yield tmp;
      tmp.length = 0;
      i = 0;
    } else {
      i++;
    }
  }
  if (tmp.length) {
    yield tmp;
  }
}
function rootRelativeFacadeId(facadeId, astroConfig) {
  return facadeId.slice(fileURLToPath(astroConfig.projectRoot).length);
}
function chunkIsPage(astroConfig, output, internals) {
  if (output.type !== "chunk") {
    return false;
  }
  const chunk = output;
  if (chunk.facadeModuleId) {
    const facadeToEntryId = prependForwardSlash(rootRelativeFacadeId(chunk.facadeModuleId, astroConfig));
    return internals.entrySpecifierToBundleMap.has(facadeToEntryId);
  }
  return false;
}
async function generatePages(result, opts, internals, facadeIdToPageDataMap) {
  var _a;
  info(opts.logging, null, `
${bgMagenta(black(" generating static routes "))}
`);
  const ssr = !!((_a = opts.astroConfig._ctx.adapter) == null ? void 0 : _a.serverEntrypoint);
  const outFolder = ssr ? getServerRoot(opts.astroConfig) : getOutRoot(opts.astroConfig);
  const ssrEntryURL = new URL(`./entry.mjs?time=${Date.now()}`, outFolder);
  const ssrEntry = await import(ssrEntryURL.toString());
  for (const pageData of eachPageData(internals)) {
    await generatePage(opts, internals, pageData, ssrEntry);
  }
}
async function generatePage(opts, internals, pageData, ssrEntry) {
  let timeStart = performance.now();
  const renderers = ssrEntry.renderers;
  const pageInfo = getPageDataByComponent(internals, pageData.route.component);
  const linkIds = Array.from((pageInfo == null ? void 0 : pageInfo.css) ?? []);
  const hoistedId = (pageInfo == null ? void 0 : pageInfo.hoistedScript) ?? null;
  const pageModule = ssrEntry.pageMap.get(pageData.component);
  if (!pageModule) {
    throw new Error(`Unable to find the module for ${pageData.component}. This is unexpected and likely a bug in Astro, please report.`);
  }
  const generationOptions = {
    pageData,
    internals,
    linkIds,
    hoistedId,
    mod: pageModule,
    renderers
  };
  const icon = pageData.route.type === "page" ? cyan("</>") : magenta("{-}");
  info(opts.logging, null, `${icon} ${pageData.route.component}`);
  const renderPromises = [];
  for (const paths of throttle(MAX_CONCURRENT_RENDERS, pageData.paths)) {
    for (const path of paths) {
      renderPromises.push(generatePath(path, opts, generationOptions));
    }
    await Promise.all(renderPromises);
    const timeEnd = performance.now();
    const timeChange = getTimeStat(timeStart, timeEnd);
    let shouldLogTimeChange = !getTimeStat(timeStart, timeEnd).startsWith("0");
    for (const path of paths) {
      const timeIncrease = shouldLogTimeChange ? ` ${dim(`+${timeChange}`)}` : "";
      info(opts.logging, null, `    ${dim("\u2503")} ${path}${timeIncrease}`);
      shouldLogTimeChange = false;
    }
    timeStart = performance.now();
    renderPromises.length = 0;
  }
}
function addPageName(pathname, opts) {
  opts.pageNames.push(pathname.replace(/\/?$/, "/").replace(/^\//, ""));
}
async function generatePath(pathname, opts, gopts) {
  const { astroConfig, logging, origin, routeCache } = opts;
  const { mod, internals, linkIds, hoistedId, pageData, renderers } = gopts;
  if (pageData.route.type === "page") {
    addPageName(pathname, opts);
  }
  debug("build", `Generating: ${pathname}`);
  const site = astroConfig.buildOptions.site;
  const links = createLinkStylesheetElementSet(linkIds.reverse(), site);
  const scripts = createModuleScriptElementWithSrcSet(hoistedId ? [hoistedId] : [], site);
  for (const script of astroConfig._ctx.scripts) {
    if (script.stage === "head-inline") {
      scripts.add({
        props: {},
        children: script.content
      });
    }
  }
  try {
    const options = {
      legacyBuild: false,
      links,
      logging,
      markdownRender: astroConfig.markdownOptions.render,
      mod,
      origin,
      pathname,
      scripts,
      renderers,
      async resolve(specifier) {
        const hashedFilePath = internals.entrySpecifierToBundleMap.get(specifier);
        if (typeof hashedFilePath !== "string") {
          if (specifier === BEFORE_HYDRATION_SCRIPT_ID) {
            return "data:text/javascript;charset=utf-8,//[no before-hydration script]";
          }
          throw new Error(`Cannot find the built path for ${specifier}`);
        }
        const relPath = npath.posix.relative(pathname, "/" + hashedFilePath);
        const fullyRelativePath = relPath[0] === "." ? relPath : "./" + relPath;
        return fullyRelativePath;
      },
      method: "GET",
      headers: new Headers(),
      route: pageData.route,
      routeCache,
      site: astroConfig.buildOptions.site,
      ssr: opts.astroConfig.buildOptions.experimentalSsr
    };
    let body;
    if (pageData.route.type === "endpoint") {
      const result = await callEndpoint(mod, options);
      if (result.type === "response") {
        throw new Error(`Returning a Response from an endpoint is not supported in SSG mode.`);
      }
      body = result.body;
    } else {
      const result = await render(options);
      if (result.type !== "html") {
        return;
      }
      body = result.html;
    }
    const outFolder = getOutFolder(astroConfig, pathname, pageData.route.type);
    const outFile = getOutFile(astroConfig, outFolder, pathname, pageData.route.type);
    await fs.promises.mkdir(outFolder, { recursive: true });
    await fs.promises.writeFile(outFile, body, "utf-8");
  } catch (err) {
    error(opts.logging, "build", `Error rendering:`, err);
  }
}
export {
  chunkIsPage,
  generatePages,
  rootRelativeFacadeId
};
